import pandas as pd
import numpy as np
from .moving_average import MovingAverage

ma = MovingAverage()


class CCI:
    """
    A class for calculating Commodity Channel Index (CCI)
    of time series data.

    Attributes:
    -----------
    source_arr : numpy.ndarray
        The input time series data as a NumPy array.
    source : pd.Series
        The input time series data as a DataFrame.
    length : int
        The number of periods to include in the CCI calculation.

    Methods:
    --------
    CCI_precise(smooth_column: str = "sma", constant: float = 0.015) -> CCI:
        Calculate CCI using a precise method.

    set_sma() -> CCI:
        Set the Simple Moving Average (SMA) for the CCI calculation.

    set_ema() -> CCI:
        Set the Exponential Moving Average (EMA) for the CCI calculation.

    CCI(constant: float = 0.015) -> pd.DataFrame:
        Calculate CCI using the specified method.
    """

    def __init__(self, source: pd.Series, length: int = 20):
        """
        Initialize the CCI object.

        Parameters:
        -----------
        source : pd.Series
            The input time series data.
        length : int, optional
            The number of periods to include in the CCI calculation, by default 20.
        """
        self.source_arr = np.array(source)
        self.source = source
        self.length = length

    def CCI_precise(
        self,
        smooth_column: str = "sma",
        constant: float = 0.015,
    ) -> pd.Series:
        """
        Calculate CCI using a precise method.
        this version have more similar results from excel
        than the standard version and TA-lib.

        Parameters:
        -----------
        smooth_column : str, optional
            The column to use for smoothing, by default "sma".
        constant : float, optional
            The constant factor for CCI calculation, by default 0.015.

        Returns:
        --------
        CCI
            The CCI object.
        """

        self.df = pd.DataFrame()
        self.df["TP"] = self.source_arr
        self.df["sma"] = self.df["TP"].rolling(self.length).mean()

        self.df["mad"] = (
            self.df["TP"]
            .rolling(self.length)
            .apply(lambda x: (pd.Series(x) - pd.Series(x).mean()).abs().mean())
        )

        self.df["CCI"] = (
            (self.df["TP"] - self.df[smooth_column])
            / (constant * self.df["mad"])
        )

        return self.df["CCI"].dropna(axis=0, inplace=True)

    def set_sma(self):
        """
        Set the Simple Moving Average (SMA) for the CCI calculation.

        Returns:
        --------
        CCI
            The CCI object.
        """
        self.window = np.ones(self.length) / self.length
        self.ma = np.convolve(self.source_arr, self.window, mode="valid")

        return self

    def set_ema(self):
        """
        Set the Exponential Moving Average (EMA) for the CCI calculation.

        Returns:
        --------
        CCI
            The CCI object.
        """
        self.ma = ma.ema(self.source, self.length).to_numpy()

        return self

    def CCI(self, constant: float = 0.015) -> pd.DataFrame:
        """
        Calculate CCI using the specified method.

        Parameters:
        -----------
        constant : float, optional
            The constant factor for CCI calculation, by default 0.015.

        Returns:
        --------
        pd.DataFrame
            The calculated CCI data as a DataFrame.
        """
        self.window = np.lib.stride_tricks.sliding_window_view(self.source_arr, self.length)
        self.mean_window = np.mean(self.window, axis=1)
        self.abs_diff = np.abs(self.window - self.mean_window[:, np.newaxis])

        self.mad = np.mean(self.abs_diff, axis=1)

        self.df = pd.DataFrame()
        self.df["source"] = self.source[self.length - 1 :]
        self.df["mad"] = self.mad
        self.df["ma"] = self.ma
        self.df["CCI"] = (
            (self.df["source"] - self.df["ma"])
            / (constant * self.df["mad"])
        )

        return self.df



from typing import Literal
import pandas as pd
import numpy as np

class MovingAverage:
    """
    A class for calculating Simple Moving Average (SMA)
    and Exponential Moving Average (EMA) of time series data.

    Attributes:
    -----------
    None

    Methods:
    --------
    sma(source: pd.Series, length: int) -> pd.Series:
        Calculate the Simple Moving Average (SMA)
        of the input time series data.

    ema(source: pd.Series, length: int) -> pd.Series:
        Calculate the Exponential Moving Average (EMA)
        of the input time series data.

    sema(source: pd.Series, length: int, smooth: int) -> pd.Series:
        Calculate the Smoothed Exponential Moving Average (SEMA)
        of the input time series data.
    """

    def sma(self, source: pd.Series, length: int) -> pd.Series:
        """
        Calculate the Simple Moving Average (SMA)
        of the input time series data.

        Parameters:
        -----------
        source : pd.Series
            The time series data to calculate the SMA for.
        length : int
            The number of periods to include in the SMA calculation.

        Returns:
        --------
        pd.Series
            The calculated SMA time series data.
        """
        sma = source.rolling(length).mean()
        return sma.dropna(axis=0)

    def ema(self, source: pd.Series, length: int) -> pd.Series:
        """
        Calculate the Exponential Moving Average (EMA)
        of the input time series data.

        Parameters:
        -----------
        source : pandas.Series
            The time series data to calculate the EMA for.
        length : int
            The number of periods to include in the EMA calculation.

        Returns:
        --------
        pandas.Series
            The calculated EMA time series data.
        """
        sma = source.rolling(window=length, min_periods=length).mean()[:length]
        rest = source[length:]
        return (
            pd.concat([sma, rest])
            .ewm(span=length, adjust=False)
            .mean()
            .dropna(axis=0)
        )

    def sema(self, source: pd.Series, length: int, smooth: int) -> pd.Series:
        """
        Calculate the Smoothed Exponential Moving Average (SEMA)
        of the input time series data.

        Parameters:
        -----------
        source : pandas.Series
            The time series data to calculate the SEMA for.
        length : int
            The number of periods to include in the SEMA calculation.
        smooth : int
            The smooth of EMAs to calculate.

        Returns:
        --------
        pandas.Series
            The calculeted SEMA time series data.
        """

        emas_dict = {}
        emas_dict["source_1"] = self.ema(source, length)
        for value in range(2, smooth + 1):
            emas_dict[f"source_{value}"] = self.ema(
                emas_dict[f"source_{value-1}"],
                length,
            )
        emas_df = pd.DataFrame(emas_dict)
        emas_df["sema"] = (
            emas_df[emas_df.columns[:-1]].diff(axis=1).sum(axis=1) * - 1
            * smooth
            + emas_df[emas_df.columns[-1]]
        )
        sema = emas_df["sema"]
        return sema.dropna(axis=0)

    def _rma_pandas(
        self,
        source: pd.Series,
        length: int,
        **kwargs
    ) -> pd.Series:
        """
        Calculate the Relative Moving Average (RMA) of the input time series
        data.

        Parameters:
        -----------
        source : pandas.Series
            The time series data to calculate the RMA for.
        length : int
            The number of periods to include in the RMA calculation.
        **kwargs : additional keyword arguments
            Additional keyword arguments to pass to the pandas EWM (Exponential
            Weighted Moving Average) function.

        Returns:
        --------
        pandas.Series
            The calculated RMA time series data.

        Note:
        -----
        The first values are different from the TradingView RMA.
        """
        sma = source.rolling(window=length, min_periods=length).mean()[:length]
        rest = source[length:]
        return (
            pd.concat([sma, rest])
            .ewm(alpha=1 / length, **kwargs)
            .mean()
        ).rename("RMA")

    def _rma_python(
        self,
        source: pd.Series,
        length: int
    ) -> pd.Series:
        """
        Calculate the Relative Moving Average (RMA) of the input time series
        data using pure python.

        Parameters:
        -----------
        source : pandas.Series
            The time series data to calculate the RMA for.
        length : int
            The number of periods to include in the RMA calculation.

        Returns:
        --------
        pd.Series
            The calculated RMA time series data.

        Note:
        -----
        The pure python version is the only one with precision in the
        initial RMA values. However, with the simple RMA version,
        both pandas and python versions will yield the same precision
        in initial values.
        """
        alpha = 1 / length
        source_pd = self._rma_pandas(source, length)[:length]
        source_values = source[length:].to_numpy().tolist()

        rma = float(source_pd.dropna().iloc[0])
        rma_list = [rma]

        for source_value in source_values:
            rma = alpha * source_value + (1 - alpha) * rma
            rma_list.append(rma)

        rma_series = pd.Series(
            rma_list,
            name="RMA",
            index=source[length - 1:].index
        )

        return rma_series

    def rma(
        self,
        source: pd.Series,
        length: int,
        method: Literal["numpy", "pandas"] = "numpy"
    ) -> np.ndarray | pd.Series:
        """
        Calculate the Relative Moving Average (RMA) of the input time series
        data.

        Parameters:
        -----------
        source : pandas.Series
            The time series data to calculate the RMA for.
        length : int
            The number of periods to include in the RMA calculation.
        method : {"numpy", "pandas"}, optional
            The method to use for calculating the RMA, by default "numpy".

        Returns:
        --------
        np.ndarray or pandas.Series
            The calculated RMA time series data.
        """
        match method:
            case "numpy":
                return self._rma_python(source, length)
            case "pandas":
                return self._rma_pandas(source, length)
            case _:
                raise TypeError("method must be 'numpy' or 'pandas'")



import pandas as pd

def OHLC_finder(
    dataframe: pd.DataFrame,
    Open: str = None,
    High: str = None,
    Low: str = None,
    Close: str = None,
) -> None:
    """
    Initialize the DMI object with the given data and
    parameters.

    Parameters:
    -----------
    dataframe : pd.DataFrame
        The DataFrame containing the source, high, and low data.
    high : str, optional
        The column name in the DataFrame representing the high
        data. If not provided, it will be inferred from common
        column names.
    low : str, optional
        The column name in the DataFrame representing the low
        data. If not provided, it will be inferred from common
        column names.
    close: str, optional
        The column name in the DataFrame representing the close
        data. If not provided, it will be inferred from common
        column names.
    """
    if not isinstance(dataframe, pd.DataFrame):
        raise ValueError("dataframe param must be a DataFrame")

    ohlc_columns_lowercase = ['open', 'high', 'low', 'close']
    ohlc_columns_uppercase = ['Open', 'High', 'Low', 'Close']

    is_ohlc_columns_lowercase = all(
        column in dataframe.columns
        for column in ohlc_columns_lowercase
    )

    is_ohlc_columns_uppercase = all(
        column in dataframe.columns
        for column in ohlc_columns_uppercase
    )

    if not is_ohlc_columns_lowercase and not is_ohlc_columns_uppercase:
        raise ValueError("OHLC columns not found in dataframe")

    if Open is None:
        if "Open" in dataframe.columns:
            open_price = dataframe["Open"]
        elif "open" in dataframe.columns:
            open_price = dataframe["open"]
        else:
            open_price = None
    else:
        open_price = dataframe[Open]

    if High is None:
        if "High" in dataframe.columns:
            high_price = dataframe["High"]
        elif "high" in dataframe.columns:
            high_price = dataframe["high"]
        else:
            high_price = None
    else:
        high_price = dataframe[High]

    if Low is None:
        if "Low" in dataframe.columns:
            low_price = dataframe["Low"]
        elif "low" in dataframe.columns:
            low_price = dataframe["low"]
        else:
            low_price = None
    else:
        low_price = dataframe[Low]

    if Close is None:
        if "Close" in dataframe.columns:
            close_price = dataframe["Close"]
        elif "close" in dataframe.columns:
            close_price = dataframe["close"]
        else:
            close_price = None
    else:
        close_price = dataframe[Close]

    return (open_price, high_price, low_price, close_price)



import numpy as np
import pandas as pd
from .moving_average import MovingAverage

ma = MovingAverage()


def RSI(source: pd.Series, periods: int=14) -> pd.Series:
    """
    Calculate the Relative Strength Index (RSI) for a given time series data.

    Parameters:
    -----------
    source : pd.Series
        The input time series data for which to calculate RSI.
    periods : int, optional
        The number of periods to use for RSI calculation, by default 14.

    Returns:
    --------
    pd.Series
        The calculated RSI values for the input data.
    """
    upward_diff = pd.Series(np.maximum(source - source.shift(1), 0.0)).dropna()

    downward_diff = pd.Series(
        np.maximum(source.shift(1) - source, 0.0)
    ).dropna()

    relative_strength = (
        ma.rma(upward_diff, periods)
        / ma.rma(downward_diff, periods)
    )

    rsi = 100 - (100 / (1 + relative_strength))
    return rsi.rename("RSI")
